/*
 * Copyright (c) 2018 Koji Suzuki
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <string.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/eeprom.h>
#include <avr/wdt.h>
#include <util/delay.h>

#define bit_set(port,bit)	{ port |= (1<<(bit));}
#define bit_clear(port,bit)	{ port &= ~(1<<(bit));}
#define min(a,b)		((a)<(b) ? (a):(b))

#define nop()	__asm__ volatile ("nop" )
#define sleep()	__asm__ volatile ("sleep" )
//#define sei()	__asm__ volatile ("sei" )
//#define cli()	__asm__ volatile ("cli" )

#ifndef TIFR0
#define TIFR0	TIFR
#endif
#ifndef TIMSK0
#define TIMSK0	TIMSK
#endif


#include "i2ckbd.h"

/*
   COL1 COL2 COL3 COL4  COL5 COL6 COL7  ROW1 ROW2 ROW3 ROW4  ROW5 ROW6 ROW7 ROW8
Type 2313
   PB6  PB4  PB3  PB2   PB1  PB0  PD6   PD0  PD1  PA1  PA0   PD5  PD4  PD3  PD2
Type t88a
   PA1  PC0  PC1  PC2   PD3  PA3  PB7   PC3  PD0  PD1  PA2   PD5  PD6  PD7  PB0
Type t88b
   PB0  PD7  PD6  PD5   PD1  PD0  PC3   PB7  PA3  PA2  PD3   PC2  PC1  PC0  PA1
Type t88c
   PC0  PC1  PC2  PC3   PD5  PD6  PD7   PD0  PD1  PD3  PB7   PB0  PB1  PB3  PB5
Type t88d,t88e
   PB5  PB4  PB0  PD7   PD1  PD0  PC3   PD6  PD5  PB7  PD3   PC2  PC1  PC0  PB1
*/

static inline void port_init() {
	GPIOR0 = 0; // flags
#if (KBD_TYPE == TYPE_T88D) || (KBD_TYPE == TYPE_T88E)
	DDRA = DDRB = DDRC = DDRD = 0;
	PORTB = (1<<PB1)| (1<<PB7); // pullup
	PORTC = (1<<PC0)| (1<<PC1) | (1<<PC2); // pullup
	PORTD = (1<<PD3)| (1<<PD5) | (1<<PD6) | (1<<PD7); // pullup
#elif (KBD_TYPE == TYPE_T88C)
	DDRA = DDRB = DDRC = DDRD = 0;
	PORTB = (1<<PB0)| (1<<PB1)| (1<<PB3)
		 (1<<PB5)| (1<<PB7); // pullup
	PORTD = (1<<PD0)| (1<<PD1) | (1<<PD3); // pullup
#elif (KBD_TYPE == TYPE_T88B)
	DDRA = DDRB = DDRC = DDRD = 0;
	PORTA = (1<<PA1)| (1<<PA2)| (1<<PA3) // pullup
	PORTB = (1<<PB7); // pullup
	PORTC = (1<<PC0)| (1<<PC1) | (1<<PC2); // pullup
	PORTD = (1<<PD3); // pullup
#elif (KBD_TYPE == TYPE_T88A)
	DDRA = DDRB = DDRC = DDRD = 0;
	PORTA = (1<<PA2); // pullup
	PORTB = (1<<PB0); // pullup
	PORTC = (1<<PC3); // pullup
	PORTD = (1<<PD0) | (1<<PD1) | (1<<PD5) 
		| (1<<PD6) | (1<<PD7); // pullup
#elif (KBD_TYPE == TYPE_861)
	DDRB = 0;
	DDRA = 0x77;
	PORTA = 0x80; // pullup
#elif (KBD_TYPE == TYPE_2313)
	DDRA = DDRB = DDRD = 0;
	PORTA = (1<<PA0)| (1<<PA1); // pullup
	PORTD = (1<<PD0) | (1<<PD1) | (1<<PD2)
		| (1<<PD3) | (1<<PD4) | (1<<PD5); // pullup
#endif
}

static inline void timer_init() {
#if (KBD_TYPE == TYPE_861)
	TCCR0A = 1; // CTC
	TCCR0B = 2 ; // 1/8 
#elif (KBD_TYPE == TYPE_2313)
	TCCR0A = 2; // CTC
	TCCR0B = 2 ; // 1/8 
#else
	TCCR0A = 2 | (1<<CTC0); // 1/8 CTC
#endif
	OCR0A = F_CPU/8/10000; // 10ms
}

static inline void enable_timer_int() {
	TIMSK0 |= (1<< OCIE0A); // enable interrupt
}
static inline int8_t timer_int_enabled() {
	return (TIMSK0 & (1<< OCIE0A));
}

#if (KBD_TYPE == TYPE_T88D) || (KBD_TYPE == TYPE_T88E)
	#define ROW1_is_clear	!(PIND & (1<<PD6))
	#define ROW2_is_clear	!(PIND & (1<<PD5))
	#define ROW3_is_clear	!(PINB & (1<<PB7))
	#define ROW4_is_clear	!(PIND & (1<<PD3))
	#define ROW5_is_clear	!(PINC & (1<<PC2))
	#define ROW6_is_clear	!(PINC & (1<<PC1))
	#define ROW7_is_clear	!(PINC & (1<<PC0))
	#define ROW8_is_clear	!(PINB & (1<<PB1))


	#define COL1_low()	{DDRB |= (1<<PB5);}
	#define COL2_low()	{DDRB |= (1<<PB4);}
	#define COL3_low()	{DDRB |= (1<<PB0);}
	#define COL4_low()	{DDRD |= (1<<PD7);}
	#define COL5_low()	{DDRD |= (1<<PD1);}
	#define COL6_low()	{DDRD |= (1<<PD0);}
	#define COL7_low()	{DDRC |= (1<<PC3);}


	#define COL1_hiz()	{DDRB &= ~(1<<PB5);}
	#define COL2_hiz()	{DDRB &= ~(1<<PB4);}
	#define COL3_hiz()	{DDRB &= ~(1<<PB0);}
	#define COL4_hiz()	{DDRD &= ~(1<<PD7);}
	#define COL5_hiz()	{DDRD &= ~(1<<PD1);}
	#define COL6_hiz()	{DDRD &= ~(1<<PD0);}
	#define COL7_hiz()	{DDRC &= ~(1<<PC3);}

#elif (KBD_TYPE == TYPE_T88C)
	#define ROW1_is_clear	!(PIND & (1<<PD0))
	#define ROW2_is_clear	!(PIND & (1<<PD1))
	#define ROW3_is_clear	!(PIND & (1<<PD3))
	#define ROW4_is_clear	!(PINB & (1<<PB7))
	#define ROW5_is_clear	!(PINB & (1<<PB0))
	#define ROW6_is_clear	!(PINB & (1<<PB1))
	#define ROW7_is_clear	!(PINB & (1<<PB3))
	#define ROW8_is_clear	!(PINB & (1<<PB5))

	#define COL1_low()	{DDRC |= (1<<PC0);}
	#define COL2_low()	{DDRC |= (1<<PC1);}
	#define COL3_low()	{DDRC |= (1<<PC2);}
	#define COL4_low()	{DDRC |= (1<<PC3);}
	#define COL5_low()	{DDRD |= (1<<PD5);}
	#define COL6_low()	{DDRD |= (1<<PD6);}
	#define COL7_low()	{DDRD |= (1<<PD7);}


	#define COL1_hiz()	{DDRC &= ~(1<<PC0);}
	#define COL2_hiz()	{DDRC &= ~(1<<PC1);}
	#define COL3_hiz()	{DDRC &= ~(1<<PC2);}
	#define COL4_hiz()	{DDRC &= ~(1<<PC3);}
	#define COL5_hiz()	{DDRD &= ~(1<<PD5);}
	#define COL6_hiz()	{DDRD &= ~(1<<PD6);}
	#define COL7_hiz()	{DDRD &= ~(1<<PD7);}

#elif (KBD_TYPE == TYPE_T88B)
	#define ROW1_is_clear	!(PINB & (1<<PB7))
	#define ROW2_is_clear	!(PINA & (1<<PA3))
	#define ROW3_is_clear	!(PINA & (1<<PA2))
	#define ROW4_is_clear	!(PIND & (1<<PD3))
	#define ROW5_is_clear	!(PINC & (1<<PC2))
	#define ROW6_is_clear	!(PINC & (1<<PC1))
	#define ROW7_is_clear	!(PINC & (1<<PC0))
	#define ROW8_is_clear	!(PINA & (1<<PA1))


	#define COL1_low()	{DDRB |= (1<<PB0);}
	#define COL2_low()	{DDRD |= (1<<PD7);}
	#define COL3_low()	{DDRD |= (1<<PD6);}
	#define COL4_low()	{DDRD |= (1<<PD5);}
	#define COL5_low()	{DDRD |= (1<<PD1);}
	#define COL6_low()	{DDRD |= (1<<PD0);}
	#define COL7_low()	{DDRC |= (1<<PC3);}


	#define COL1_hiz()	{DDRB &= ~(1<<PB0);}
	#define COL2_hiz()	{DDRD &= ~(1<<PD7);}
	#define COL3_hiz()	{DDRD &= ~(1<<PD6);}
	#define COL4_hiz()	{DDRD &= ~(1<<PD5);}
	#define COL5_hiz()	{DDRD &= ~(1<<PD1);}
	#define COL6_hiz()	{DDRD &= ~(1<<PD0);}
	#define COL7_hiz()	{DDRC &= ~(1<<PC3);}

#elif (KBD_TYPE == TYPE_T88A)
	#define ROW1_is_clear	!(PINC & (1<<PC3))
	#define ROW2_is_clear	!(PIND & (1<<PD0))
	#define ROW3_is_clear	!(PIND & (1<<PD1))
	#define ROW4_is_clear	!(PINA & (1<<PA2))
	#define ROW5_is_clear	!(PIND & (1<<PD5))
	#define ROW6_is_clear	!(PIND & (1<<PD6))
	#define ROW7_is_clear	!(PIND & (1<<PD7))
	#define ROW8_is_clear	!(PINB & (1<<PB0))

	#define COL1_low()	{DDRA |= (1<<PA1);}
	#define COL2_low()	{DDRC |= (1<<PC0);}
	#define COL3_low()	{DDRC |= (1<<PC1);}
	#define COL4_low()	{DDRC |= (1<<PC2);}
	#define COL5_low()	{DDRD |= (1<<PD3);}
	#define COL6_low()	{DDRA |= (1<<PA3);}
	#define COL7_low()	{DDRB |= (1<<PB7);}


	#define COL1_hiz()	{DDRA &= ~(1<<PA1);}
	#define COL2_hiz()	{DDRC &= ~(1<<PC0);}
	#define COL3_hiz()	{DDRC &= ~(1<<PC1);}
	#define COL4_hiz()	{DDRC &= ~(1<<PC2);}
	#define COL5_hiz()	{DDRD &= ~(1<<PD3);}
	#define COL6_hiz()	{DDRA &= ~(1<<PA3);}
	#define COL7_hiz()	{DDRB &= ~(1<<PB7);}

#elif (KBD_TYPE == TYPE_2313)

	#define ROW1_is_clear	!(PIND & (1<<PD0))
	#define ROW2_is_clear	!(PIND & (1<<PD1))
	#define ROW3_is_clear	!(PINA & (1<<PA1))
	#define ROW4_is_clear	!(PINA & (1<<PA0))
	#define ROW5_is_clear	!(PIND & (1<<PD5))
	#define ROW6_is_clear	!(PIND & (1<<PD4))
	#define ROW7_is_clear	!(PIND & (1<<PD3))
	#define ROW8_is_clear	!(PIND & (1<<PD2))

	#define COL1_low()	{DDRB |= (1<<PB6);}
	#define COL2_low()	{DDRB |= (1<<PB4);}
	#define COL3_low()	{DDRB |= (1<<PB3);}
	#define COL4_low()	{DDRB |= (1<<PB2);}
	#define COL5_low()	{DDRB |= (1<<PB1);}
	#define COL6_low()	{DDRB |= (1<<PB0);}
	#define COL7_low()	{DDRD |= (1<<PD6);}

	#define COL1_hiz()	{DDRB &= ~(1<<PB6);}
	#define COL2_hiz()	{DDRB &= ~(1<<PB4);}
	#define COL3_hiz()	{DDRB &= ~(1<<PB3);}
	#define COL4_hiz()	{DDRB &= ~(1<<PB2);}
	#define COL5_hiz()	{DDRB &= ~(1<<PB1);}
	#define COL6_hiz()	{DDRB &= ~(1<<PB0);}
	#define COL7_hiz()	{DDRD &= ~(1<<PD6);}
#endif

#if (KBD_TYPE == TYPE_861)
static inline void __keyscan(uint8_t *r) {
	uint8_t col,row,a;
	uint8_t msk;
	for (col=0; col<7; col++) {
	    PORTA = (PORTA & ~0x70) | ((col<<4) & 0x70);
	    a = 0;
	    msk = 0x01;
	    for (row=0; row<7; row++) {
	        PORTA = (PORTA & ~0x07) | (row & 0x07);
		nop();
		nop();
		if (!(PINA & (1<<PA7))) a |= msk;
	        msk <<= 1;
	    }
	    *r++ = a;
	}
	PORTA = PORTA | 0x70;
}
#else
static inline uint8_t scanline() {
	uint8_t r = 0;
	if (ROW1_is_clear) r |= (1<<0);
	if (ROW2_is_clear) r |= (1<<1);
	if (ROW3_is_clear) r |= (1<<2);
	if (ROW4_is_clear) r |= (1<<3);
	if (ROW5_is_clear) r |= (1<<4);
	if (ROW6_is_clear) r |= (1<<5);
	if (ROW7_is_clear) r |= (1<<6);
	if (ROW8_is_clear) r |= (1<<7);
	return r;
}

 #ifdef USE_GPIOR1
static inline void __keyscan(uint8_t *r) {
	for (GPIOR1=1; GPIOR1 != (1<<7); GPIOR1 <<=1) {
		if (GPIOR1 & (1<<0)) COL1_low();
		if (GPIOR1 & (1<<1)) COL2_low();
		if (GPIOR1 & (1<<2)) COL3_low();
		if (GPIOR1 & (1<<3)) COL4_low();
		if (GPIOR1 & (1<<4)) COL5_low();
		if (GPIOR1 & (1<<5)) COL6_low();
		if (GPIOR1 & (1<<6)) COL7_low();

		nop();
		nop();
		*r++ = scanline();

		if (GPIOR1 & (1<<0)) COL1_hiz();
		if (GPIOR1 & (1<<1)) COL2_hiz();
		if (GPIOR1 & (1<<2)) COL3_hiz();
		if (GPIOR1 & (1<<3)) COL4_hiz();
		if (GPIOR1 & (1<<4)) COL5_hiz();
		if (GPIOR1 & (1<<5)) COL6_hiz();
		if (GPIOR1 & (1<<6)) COL7_hiz();
	}
}
 #else
static inline void __keyscan(uint8_t *r) {
	int8_t col;
	for (col=0; col<7; col++) {
		switch (col) {
		    case 0: COL1_low(); break;
		    case 1: COL2_low(); break;
		    case 2: COL3_low(); break;
		    case 3: COL4_low(); break;
		    case 4: COL5_low(); break;
		    case 5: COL6_low(); break;
		    case 6: COL7_low(); break;
		    default: break;
		}
		nop();
		nop();
		*r++ = scanline();
		switch (col) {
		    case 0: COL1_hiz(); break;
		    case 1: COL2_hiz(); break;
		    case 2: COL3_hiz(); break;
		    case 3: COL4_hiz(); break;
		    case 4: COL5_hiz(); break;
		    case 5: COL6_hiz(); break;
		    case 6: COL7_hiz(); break;
		    default: break;
		}
	}
}
 #endif
#endif

#ifdef FN_EXTEND
uint8_t keystat_tab[7*3];
#else
uint8_t keystat_tab[7];
#endif

static inline uint8_t key_pressed(uint8_t code) {
	return keystat_tab[code >>3] & (1<<(code & 0x7));
}

uint8_t read_buf_len;
uint8_t read_buf[10];

static inline void kbd_putc(uint8_t code) {
	if (read_buf_len < 10) 
	    read_buf[read_buf_len++] = code;
}

#define LCODE_CTRL	2
#define LCODE_SHIFT	15
#define LCODE_ALT	3
#define LCODE_FN	53

#ifdef SUPPORT_ASCIICODE
const uint8_t ascii_tab[56*2] PROGMEM = {
'1', 0x09, 0xf0, 0xf2 , 0x08,  ']', 0x0d,  0x1b,
'2', 'q',  'a',  'z',   '=',  '[',  '\'',  0xf1,
'3', 'w',  's',  'x',   '-',  'p',  ':',   '/',
'4', 'e',  'd',  'c',   '0',  'o',  'l',   '.',
'5', 'r',  'f',  'v',   '9',  'i',  'k',   ',',
'6', 't',  'g',  'b',   '8',  'u',  'j',   'm',
'7', 'y',  'h',  'n',   '\\', 0xff, ' ',   ' ',
/* SHIFT */
'!', 0x09, 0xf0, 0xf2 , 0x08,  '}', 0x0d,  0x1b,
'@', 'Q',  'A',  'Z',   '+',  '{',  '"',   0xf1,
'#', 'W',  'S',  'X',   '_',  'P',  ';',   '?',
'$', 'E',  'D',  'C',   ')',  'O',  'I',   '<', 
'%', 'R',  'F',  'V',   '(',  'I',  'K',   '>',
'^', 'T',  'G',  'B',   '*',  'U',  'J',   'M',
'&', 'Y',  'H',  'N',   '|',  0xff, ' ',   ' '
};
#endif

#ifdef SUPPORT_SCANCODE1
const uint8_t scancode1_tab[56*2] PROGMEM = {
0x02, 0x0F, 0x0F, 0x38,   0x0E, 0x1B, 0x1C,  0x01,
0x03, 0x10, 0x1E, 0x2C,   0x0D, 0x1A, 0x28,  0x2A,
0x04, 0x11, 0x1F, 0x2D,   0x0C, 0x19, 0x27,  0x35,
0x05, 0x12, 0x20, 0x2E,   0x0B, 0x18, 0x26,  0x34,
0x06, 0x13, 0x21, 0x2F,   0x0A, 0x17, 0x25,  0x33,
0x07, 0x14, 0x22, 0x30,   0x09, 0x16, 0x24,  0x32,
0x08, 0x15, 0x23, 0x31,   0x7D, 0xff, 0x39,  0x39,
/* FN */
0x3B, 0x3A, 0xff, 0xff,   0xD3, 0x7D, 0xff,  0x29,
0x3C, 0x01, 0xff, 0xff,   0xD2, 0xD2, 0xCD,  0xff,
0x3D, 0x29, 0xff, 0xff,   0x57, 0xff, 0xCB,  0xD0,
0x3E, 0xff, 0xff, 0xff,   0x44, 0x54, 0xC9,  0xD1,
0x3F, 0xff, 0xff, 0xff,   0x43, 0xff, 0xff,  0xff,
0x40, 0xff, 0xff, 0xff,   0x42, 0xff, 0xff,  0xff,
0x41, 0xff, 0xff, 0xff,   0xff, 0xff, 0xff,  0xff
};
#endif

#ifdef SUPPORT_HIDCODE
const uint8_t hidcode_tab[56*2] PROGMEM = {
0x1E, 0x2B, 0xE0, 0xE2,   0x2A, 0x30, 0x28, 0x29,
0x1F, 0x14, 0x04, 0x1D,   0x2E, 0x2F, 0x34, 0xE1,
0x20, 0x1A, 0x16, 0x1B,   0x2D, 0x13, 0x33, 0x38,
0x21, 0x08, 0x07, 0x06,   0x27, 0x12, 0x0F, 0x37,
0x22, 0x15, 0x09, 0x19,   0x26, 0x0C, 0x0E, 0x36,
0x23, 0x17, 0x0A, 0x05,   0x25, 0x18, 0x0D, 0x10,
0x24, 0x1C, 0x0B, 0x11,   0x89, 0xff, 0x2C, 0x2C,
/* FN */
0x3A, 0xff, 0xff, 0xff,   0x4C, 0x89, 0xff, 0x35,
0x3B, 0x29, 0xff, 0xff,   0x49, 0x52, 0x4F, 0xff,
0x3C, 0x35, 0xff, 0xff,   0x44, 0xff, 0x50, 0x51,
0x3D, 0xff, 0xff, 0xff,   0x43, 0x46, 0x4B, 0x4E,
0x3E, 0xff, 0xff, 0xff,   0x42, 0xff, 0xff, 0xff,
0x3F, 0xff, 0xff, 0xff,   0x41, 0xff, 0xff, 0xff,
0x40, 0xff, 0xff, 0xff,   0xff, 0xff, 0xff, 0xff
};
#endif

static inline void key_make(int8_t pressed, uint8_t code) {
	uint8_t mode = kbd_mode;

	if (mode == 1) { // local scancode
#ifdef SUPPORT_LOCALCODE
	    code++;
	    if (pressed) {
		kbd_putc(code);
	    } else {
		kbd_putc((code) | 0x80);
	    }
#endif
#ifdef SUPPORT_ASCIICODE
	} else if (mode == 2) { // ascii 
	    if (pressed) {
		if (code == LCODE_CTRL) {
			set_F_CTRL();
		} else if (code == LCODE_SHIFT) {
			set_F_SHIFT();
		} else if (code == LCODE_ALT) {
			set_F_ALT();
		} else {
			if (code >= 56) code -= 56;
			if (F_FN) {
			} else if (F_CTRL) {
				kbd_putc(pgm_read_byte_near(ascii_tab+code) &0x1f);
			} else if (F_SHIFT) {
				kbd_putc(pgm_read_byte_near(ascii_tab+code+56));
			} else {
				kbd_putc(pgm_read_byte_near(ascii_tab+code));
			}
		}
	    } else {
		if (code == LCODE_CTRL) {
			clear_F_CTRL();
		} else if (code == LCODE_SHIFT) {
			clear_F_SHIFT();
		} else if (code == LCODE_ALT) {
			clear_F_ALT();
		}
	    }
#endif
#ifdef SUPPORT_SCANCODE1
	} else if (mode == 3) { // scancode1
	    uint8_t code1 = pgm_read_byte_near(scancode1_tab+code);
	    if ((code1 == 0xff) && (code >= 56))
		code1 = pgm_read_byte_near(scancode1_tab+code-56);
	    if (code1 == 0xff) {
	    } else if (pressed) {
		if (code1 & 0x80) kbd_putc(0xE0);
		kbd_putc(code1 & 0x7f);
	    } else {
		if (code1 & 0x80) kbd_putc(0xE0);
		kbd_putc(code1 | 0x80);
	    }
#endif
	}
}

#ifdef SUPPORT_HIDCODE
#ifndef STACK_TUNE
static inline
#endif
void hidkey_make() {
	int8_t col,row;
	uint8_t msk;
	uint8_t *q;
	int8_t code = 0;

	kbd_putc(0x00);
	kbd_putc(0x00);
#ifdef FN_EXTEND
	q = keystat_tab +7;
	for (col=0; col<14; col++) {
#else
	q = keystat_tab;
	for (col=0; col<7; col++) {
#endif
	    msk = 1;
	    for (row=0; row<8; row++) {
		if (*q & msk) {
		    uint8_t code1 = pgm_read_byte_near(hidcode_tab+code);
		    if ((code1 == 0xff) && (code >= 56))
			code1 = pgm_read_byte_near(hidcode_tab+code-56);
		    if (code1 == 0xff) { /* FN */
		    } else if (code1 == 0xE0) { /* CTRL */
			read_buf[0] |= 0x01;
		    } else if (code1 == 0xE1) { /* ALT */
			read_buf[0] |= 0x04;
		    } else if (code1 == 0xE2) { /* SHIFT */
			read_buf[0] |= 0x02;
		    } else {
			kbd_putc(code1);
		    }
		}
		msk <<= 1;
	        code ++;
	    }
	    q++;
	}
	while (read_buf_len <8) kbd_putc(0x00);
	if (read_buf_len > 8) {
	    read_buf_len = 2;
	    while (read_buf_len <8) kbd_putc(0xff);
	}
}
#endif

static uint8_t scan_data[4*7];
static uint8_t scan_index;

#ifndef STACK_TUNE
static inline
#endif
void keystat_scan() {
	int8_t col,row;
	uint8_t r,s,msk;
	uint8_t *p,*q;
	int8_t code = 0;
	p = scan_data;
	q = keystat_tab;
	__keyscan(p + (scan_index <<3));
	for (col=0; col<7; col++) {
	    s = p[0] & p[7] & p[7*2] & p[7*3];
	    r = p[0] | p[7] | p[7*2] | p[7*3];
	    msk = 1;
	    for (row=0;row<8;row++) {
		if ((msk & s) && !(msk & *q)) {
		    *q |= msk;
#ifdef FN_EXTEND
		    if (code == LCODE_FN) {
			set_F_FN();
		    } else if ( F_FN
			&& (code != LCODE_CTRL) && (code != LCODE_SHIFT)
			&& (code != LCODE_ALT) && (code != LCODE_FN) ) {
			*(q+14) |= msk;
			key_make(1, code + row +56);
		    } else {
			*(q+7) |= msk;
			key_make(1, code + row);
		    }
#else
		    key_make(1, code + row);
#endif
		}
		if (!(msk & r) && (msk & *q)) {
		    *q &= ~msk;
#ifdef FN_EXTEND
		    if (code == LCODE_FN) {
			clear_F_FN();
		    } else {
			if (*(q+14) & msk) { // key_pressed_ex
			    *(q+14) &= ~msk;
			    key_make(0, code + row +56);
			}
			if (*(q+7) & msk) { // key_pressed_ex
			    *(q+7) &= ~msk;
			    key_make(0, code + row);
			}
		    }
#else
		    key_make(0, code + row);
#endif
		}
		msk <<=1;
	    }
	    code += 8;
	    p++;
	    q++;
	}
	scan_index = (scan_index +1) & 0x3;
}

#ifndef STACK_TUNE
static inline
#endif
void init_main() {
	port_init();
	timer_init();
	i2c_init();
#ifdef USE_SLEEP
	enable_timer_int();
#endif
}

int main() {
	init_main();

	while (1) {
	    sei();
#ifdef USE_SLEEP
	    if (timer_int_enabled()) {
		sleep();
	    }
#endif
#ifdef STACK_TUNE
	    cli();
#endif
	    if (F_STOP_SCAN) {
	    } else if (F_MODE_CHG) {
		read_buf_len = 0;
		clear_F_MODE_CHG();
	    } else if (F_TIMER_TICK0 || F_TIMER_TICK) {
		if (F_TIMER_TICK0) clear_F_TIMER_TICK0();
		if (F_TIMER_TICK) clear_F_TIMER_TICK();
		keystat_scan();
	    }
#ifdef SUPPORT_HIDCODE
	    if (F_BUILD_REQ) {
		hidkey_make();
	    	clear_F_BUILD_REQ();
	    	set_F_BUILD_DONE();
	    }
#endif
	}
}

SIGNAL(TIMER0_COMPA_vect) {
	set_F_TIMER_TICK();
}
