/*
 * Copyright (c) 2007-2010 Koji Suzuki
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/wdt.h>

#include "i2ckbd.h"
#include "usiconf.h"

#define _USIBR	USIDR

#define bit_set(port,bit)	{ port |= (1<<(bit));}
#define bit_clear(port,bit)	{ port &= ~(1<<(bit));}


volatile uint8_t kbd_mode;


const uint8_t localdesc[] PROGMEM = {
	'a', 'b', 'c', 'd',
	0x00, 0x00
};

#ifdef SUPPORT_HIDCODE
const uint8_t reportdesc[35] PROGMEM = {
    0x05, 0x01,  // USAGE_PAGE (Generic Desktop)
    0x09, 0x06,  // USAGE (Keyboard)
    0xa1, 0x01,  // COLLECTION (Application)
    0x05, 0x07,     //   USAGE_PAGE (Keyboard)
    0x19, 0xe0,     //   USAGE_MINIMUM (Keyboard LeftControl)
    0x29, 0xe7,     //   USAGE_MAXIMUM (Keyboard Right GUI)
    0x15, 0x00,     //   LOGICAL_MINIMUM (0)
    0x25, 0x01,     //   LOGICAL_MAXIMUM (1)
    0x75, 0x01,     //   REPORT_SIZE (1)
    0x95, 0x08,     //   REPORT_COUNT (8)
    0x81, 0x02,     //   INPUT (Data,Var,Abs)
    0x95, 0x01,     //   REPORT_COUNT (1)
    0x75, 0x08,     //   REPORT_SIZE (8)
    0x25, 0x65,     //   LOGICAL_MAXIMUM (101)
    0x19, 0x00,     //   USAGE_MINIMUM (Reserved (no event indicated))
    0x29, 0x65,     //   USAGE_MAXIMUM (Keyboard Application)
    0x81, 0x00,     //   INPUT (Data,Ary,Abs)
    0xc0        // END_COLLECTION };
};
const uint8_t hiddesc[28] PROGMEM = {
    0x00, 0x01,                 // Version 0x0100
    sizeof(reportdesc)+2, 0x00, // Report Decriptor Length
    REPORTDESC_REGISTER, 0x00,  // 
    INPUT_REGISTER, 0x00,       // 
    10, 0x00,                   // Max Input Length
    OUTPUT_REGISTER, 0x00,      // 
    10, 0x00,                   // Max Output Length
    COMMAND_REGISTER, 0x00,     // 
    DATA_REGISTER, 0x00,        // 
    0x05, 0x00,                 // Vendor 
    0x05, 0x00,                 // Product
    0x05, 0x00,                 // Version
    0x00, 0x00, 0x00, 0x00      // reserved :0
};
#endif

#define STATE_IDLE		0
#define STATE_RECV_ADDR		1

#define STATE_SEND_REPLY1	2
#define STATE_RECV_DATA1	3	// addr_L

#define STATE_SEND_REPLY2	4
#define STATE_RECV_DATA2	5	// addr_H

#define STATE_SEND_REPLY3	6
#define STATE_RECV_DATA3	7	// cmd

#define STATE_SEND_REPLY4	8
#define STATE_RECV_DATA4	9	// cmd L

#define STATE_SEND_REPLY	10
#define STATE_RECV_DATA		11

#define STATE_SEND_DATA1_START	12
#define STATE_SEND_DATA1	13	// length_L
#define STATE_RECV_REPLY1	14

#define STATE_SEND_DATA2_START	15
#define STATE_SEND_DATA2	16	// lenght_H

#define STATE_RECV_REPLY	17
#define STATE_SEND_DATA_START	18
#define STATE_SEND_DATA		19	// data


/* 
 * IDLE -> RECV_ADDR 
 *                  (WRITE COMMAND)
 *                   -> SEND_REPLY -> RECV_DATA -+
 *                         |                     |
 *                         +---<-----------------+
 *
 *                  (READ COMMAND)
 *                   -> SEND_DATA_START -> SEND_DATA -> RECV_REPLAY -+
 *                                             |                     |
 *                                             +---<-----------------+
 *
 * note: goto IDLE with any error /or complete
 *
 */

static uint8_t state;
static uint8_t cur_cmd;


/* 
 *          getc    return getc  1st putc   done getc     flush
 *           (A)        (B)        (C)         (D)          (E)
 * WRITE  ---------------|----------------------|---------------------
 *          Accept       |        Deny          |     Accept
 *                
 */

static inline void set_idle() {
	bit_clear(I2CDEV_DDR, I2CDEV_SDA);
	USICR = (1<<USISIE)|(2<<USIWM0)|(2<<USICS0);
	USISR = (1<<USIOIF)|(1<<USIPF)|(1<<USIDC) | (0<<USICNT0);
}

static inline void send_ack() {
	USIDR = 0;
	bit_set(I2CDEV_DDR, I2CDEV_SDA);
	USISR = (1<<USIOIF)|(1<<USIPF)|(1<<USIDC) | (14<<USICNT0);
}

static inline void recv_reply() {
	bit_clear(I2CDEV_DDR, I2CDEV_SDA);
	USIDR = 0;
	USISR = (1<<USIOIF)|(1<<USIPF)|(1<<USIDC) | (14<<USICNT0);
}

static uint8_t *rx_buf;
static int8_t rx_len;
static int8_t rx_idx;

static inline void send_data() {
	if (rx_len > rx_idx) {
	    if (F_PGM_DATA)
		USIDR = pgm_read_byte_near(rx_buf+rx_idx);
	    else 
		USIDR = rx_buf[rx_idx];
	    rx_idx++;
	} else {
		rx_len = 0; // send complete
		USIDR = 0x00;
	}
	bit_set(I2CDEV_DDR, I2CDEV_SDA);
	USISR = (1<<USIOIF)|(1<<USIPF)|(1<<USIDC) | (0<<USICNT0);
}

static inline void recv_data() {
	bit_clear(I2CDEV_DDR, I2CDEV_SDA);
	USISR = (1<<USIOIF)|(1<<USIPF)|(1<<USIDC) | (0<<USICNT0);
}


void i2c_init() {

	USICR = 0;
	bit_set(I2CDEV_PORT,I2CDEV_SCL);
	bit_set(I2CDEV_PORT,I2CDEV_SDA);
	bit_set(I2CDEV_DDR,I2CDEV_SCL);
	bit_clear(I2CDEV_DDR,I2CDEV_SDA);

	USISR = (1<<USISIF) | (1<<USIOIF) | (1<<USIPF) | (1<<USIDC);
	state = STATE_IDLE;
	/* enable I2C and USI_START interrupt enable */
	USICR = (1<<USISIE) | (2<<USIWM0) | (2<<USICS0);
}

SIGNAL(USI_START_vect) {
	bit_clear(I2CDEV_DDR, I2CDEV_SDA);

	/* wait for start condition completed :
	 *
	 * SDA ~~~~~|_________________
         * SCL ~~~~~~~~~~~~~~~~~|______
	 *          @ now       @
	 *          |<--------->|
	 */
		 
	USICR &= ~((1<<USISIE) | (1<<USIOIE));
	sei();  // enable other interrupt (except USI_OVERFLOW,USI_START)
	while (bit_is_clear(USISR,USIPF) && bit_is_set(I2CDEV_PIN, I2CDEV_SCL))
		;
	cli();
	//usi_start_cond_time = get_time(); // for sync 
	state = STATE_RECV_ADDR;

	/* enable I2C (with SCL low) 
	 *    and USI_START/USI_OVERFLOW interrupt enable
	 */ 
	/* USICR change : set USIOIE (enable Overflow interrupt)
	 *              : USIWMO 2->3 (Low SCL until clear USIOIF)
	 */
	USICR = (1<<USISIE) | (1<<USIOIE) | (3<<USIWM0) | (2<<USICS0);
	USISR =  (1<<USISIF)|(1<<USIOIF)|(1<<USIPF)|(1<<USIDC) |(0<<USICNT0);
}

SIGNAL(USI_OVERFLOW_vect) {
	int8_t data;
	switch (state) {
	case STATE_RECV_ADDR:
		if ((_USIBR == 0) || ((_USIBR & ~1) == (I2CKBD_ADDR<<1))) {
			if (_USIBR & 1) { // read command
				state = STATE_SEND_DATA_START;
			} else { // write command
				// XXX
				clear_F_STOP_SCAN();
				state = STATE_SEND_REPLY1;
			}
			send_ack();
		} else {
			goto reinit;
		}
		break;

	case STATE_SEND_REPLY1:
	case STATE_SEND_REPLY2:
	case STATE_SEND_REPLY3:
	case STATE_SEND_REPLY4:
	case STATE_SEND_REPLY:
		state ++;
		recv_data();
		rx_len = rx_idx = 0;
		break;
	case STATE_RECV_DATA1: // register L
		state = STATE_SEND_REPLY;
		data = _USIBR;
		if (data == COMMAND_REGISTER) {
			state = STATE_SEND_REPLY2;
		} else if (data == INPUT_REGISTER) {
			rx_len = -2;
			rx_idx = 0;
		} else if (data == DATA_REGISTER) {
			rx_buf = read_buf;
			//rx_len = read_buf_len;
			rx_idx = 0;
			set_F_STOP_SCAN();
			clear_F_PGM_DATA();
			
		} else if (data == KEYSTAT_REGISTER) {
#ifdef FN_EXTEND
			rx_buf = keystat_tab +7;
			rx_len = 14;
#else
			rx_buf = keystat_tab;
			rx_len = 7;
#endif
			rx_idx = 0;
			set_F_STOP_SCAN();
			clear_F_PGM_DATA();
		} else if (data == LOCALDESC_REGISTER) {
			rx_buf = (uint8_t *)localdesc;
			rx_len = sizeof(localdesc);
			set_F_PGM_DATA();
#ifdef SUPPORT_HIDCODE
		} else if (data == HIDDESC_REGISTER) {
			rx_buf = (uint8_t *)hiddesc;
			rx_len = sizeof(hiddesc);
			set_F_PGM_DATA();
		} else if (data == REPORTDESC_REGISTER) {
			rx_buf = (uint8_t *)reportdesc;
			rx_len = sizeof(reportdesc);
			set_F_PGM_DATA();
#endif
		}
		send_ack();
		break;
	case STATE_RECV_DATA2: // register H
		_USIBR;
		state = STATE_SEND_REPLY3;
		send_ack();
		break;
	case STATE_RECV_DATA3: // cmd 
		state = STATE_SEND_REPLY4;
		cur_cmd = _USIBR;
		break;
	case STATE_RECV_DATA4: // cmd data
		state = STATE_SEND_REPLY1;
		data = _USIBR;
		if ((cur_cmd & 0xf0) == CMD_VENDOR) {
			kbd_mode = data & 0x0f;
			set_F_MODE_CHG();
#ifdef SUPPORT_HIDCODE
		} else if ((cur_cmd & 0xf0) == CMD_GET_REPORT) {
			set_F_BUILD_REQ();
#endif
		}
		break;
	case STATE_RECV_DATA:
		_USIBR;
		state = STATE_SEND_REPLY;
		send_ack();
		break;

	case STATE_SEND_DATA1_START:
		if (F_BUILD_DONE) {
			clear_F_BUILD_DONE();
			rx_len = read_buf_len;
		}
		USIDR = rx_len +2; // Lehgth L
		state = STATE_SEND_DATA1;
		break;
	case STATE_SEND_DATA1:
		recv_reply();
		state = STATE_RECV_REPLY1;
		break;
	case STATE_RECV_REPLY1:
		if ( _USIBR ) { // NACK : last data
			goto reinit;
		}
		// continue to STATE_SEND_DATA2_START
	case STATE_SEND_DATA2_START:
		USIDR = 0; // Lehgth H
		state = STATE_SEND_DATA2;
		break;
	case STATE_SEND_DATA2:
		recv_reply();
		state = STATE_RECV_REPLY;
		break;
	case STATE_RECV_REPLY:
		if ( _USIBR ) { // NACK : last data
			goto reinit;
		}
		// continue to STATE_SEND_DATA_START
	case STATE_SEND_DATA_START:
		send_data(); // DATA
		state = STATE_SEND_DATA;
		break;
	case STATE_SEND_DATA:
		recv_reply();
		state = STATE_RECV_REPLY;
		break;
	}
	return;
reinit:
	rx_idx = 0;
	state = STATE_IDLE;
	set_idle();
}
